import java.math.* ;
import java.io.* ;
/*
  QS

 Luc Maranget, projet Moscova, Inria Rocquencourt
 
 Copyright 2005 Institut National de Recherche en Informatique et    
 en Automatique.  All rights reserved.  This file is distributed     
 under the terms of the Q Public License version 1.0.          
*/

class Sieve  {
  int verbose ;
  private int fc = 1 ;
  private int S ;
  private BigInteger B ;
  private BigInteger lim ;
  short logB, logMaxPrime ;
  int [] base ;
  BigInteger [] bbase ;
  int [] root ;
  private BigInteger n ;
  int sz ;
  private Table tab = new Table () ;
  short [] w ;
  private Collector col ;
  private PrintWriter log ;

  Sieve (int [] p, int [] r,  BigInteger n, int sz, int B, Collector col,
         Options o) {
    this.verbose = o.verbose ;
    // Do not sieve small primes when factor base is ``large enough''
    this.S = p.length < 100 ? 0 : o.S ;
    // if factor base is small, consider large prime variation to get more
    // equations.
    if (p.length < 50) fc = 32 ;
    if (o.F != 0) fc = o.F ;
    
    base = p ;
    bbase = new BigInteger [p.length] ;
    for (int i = 0 ; i < p.length ; i++) {
      bbase[i] = BigInteger.valueOf(p[i]) ;
    }
    root = r ;    
    this.n = n ;
    this.sz = sz ;

    logB = (short)((Math.log(B) + Math.log(fc)) / Math.log(2)) ;
    logMaxPrime = (short)(Math.log(base[base.length-1])/ Math.log(2)) ;
    if (verbose > 0){
      System.err.println("logB=" + logB + ", nprimes=" + p.length) ;
    }
    this.B = BigInteger.valueOf(B) ;
    this.lim = BigInteger.valueOf(fc).multiply(this.B) ;
    w = new short[sz] ;
    this.col = col ;
    if (verbose > 0) {
      try {
        log = new PrintWriter (new FileWriter ("log")) ;
      } catch (Exception e) {
      }
    }
  }

  BigInteger poly(BigInteger x) {
    x = x.multiply(x) ;
    x = x.subtract(n) ;
    return x ;
  }

  // Notice that factorisation produces an array of index in the base
  // Except for an additional 'small prime'
  private int [] factorThis(BigInteger y) {
    int len = base.length ;    

    if (y.signum() < 0) {
      tab.add(0) ;
      y = y.abs() ;
    }

    // Optimise the case of 2, pfou
    int k = y.getLowestSetBit() ;
    if (k > 0) {
      for (int i = 0 ; i < k ; i++) {
        tab.add(1) ;
      }
      y = y.shiftRight(k) ;
    }

    // General trial division, notice that prime indices, not primes,
    // get recorded.
    for (int i = 2 ; i < len ; i++) {
      BigInteger p = bbase [i] ;
      BigInteger [] qr = y.divideAndRemainder(p) ;
      BigInteger q = qr[0] ;
      BigInteger r = qr[1] ;      
      while (r.signum() == 0) {
        tab.add(i) ; 
        y = q ;
        qr = y.divideAndRemainder(p) ;
        q = qr[0] ;
        r = qr[1] ;
      }
      if (q.compareTo(p) < 0) { // q is prime
        if (BigInteger.ONE.equals(y)) // q is one
          return tab.trim() ;
        if (this.lim.compareTo(y) > 0) { // if q is relatively small
          if (y.bitLength() >= 30) break ;
          int lastFactor = y.intValue() ;
          for (i++ ; i < len ; i++) { // q may belong to base
            if (lastFactor == base[i]) {
              tab.add(i) ;
              return tab.trim() ;
            }
          }
          tab.add(lastFactor) ; // Q is a 'small' prime, not in the base
          return tab.trim() ;
        }
      }
    }
    // factorisation on base failed
    tab.empty() ;
    return null ;
  }

  private void efficiency (int good, int tried) {
    if (tried > 0) {
      double e = ((200 * good)/tried) / 2.0 ;
      System.err.print(" " + good + "/" + tried + " " + e) ;
    } else {
      System.err.print(" 0/0") ;
    }
  }

  void efficiency(int id) {
    if (log != null) {
      log.println(id + " " + lgood) ;
    }
    System.err.print("S" + id + ":") ;
    efficiency (lgood, ltried) ;
    good += lgood ; tried += ltried ;    
    lgood = ltried = 0 ;
    System.err.print(",") ;
    efficiency (good, tried) ;
    System.err.println() ;
  }

  void finalEfficiency(int nsteps) {
    if (log != null) {
      log.close() ;
    }
    good += lgood ; tried += ltried ;
    System.err.print("Efficiency: ") ;
    efficiency (good, tried) ;
    System.err.println() ;
  }

  // inlined later for speed
  // xi smaller
  private void justSieve(int xi, int yi, int p, int log) {
    int k = xi ;
    for (int kk=yi ; kk < sz ; k += p, kk += p) {
      w[k] -= log ;
      w[kk] -= log ;
    }
    if (k < sz)
      w[k] -= log ;
  }

  BigInteger v ; // value of x0
  private int tried,lgood ;
  int good, ltried ; 
  private final static short BIG = (short)(((short)1) << 15) ;

  boolean tryFactorThis(int k) {
    BigInteger x = v.add(BigInteger.valueOf (k)) ;
    BigInteger y = poly(x) ;
    int [] t = factorThis(y) ;
    ltried++ ;
    w[k] = BIG ;
    if (t != null) {
      lgood++ ;
      if (col.store(x, t)) {
        return true ;
      }
    }
    return false ;
  }

  // inlined later for speed
  private boolean doSieve(int xi, int yi, int p, int log) {
    int k = xi ;
    for (int kk = yi ; kk < sz ; k += p, kk += p) {
      if ((w[k] -= log) <= 0) {
        if (tryFactorThis(k)) return true ;
      }
      if ((w[kk] -= log) <= 0) {
        if (tryFactorThis(kk)) return true ;
      }
    }
    if (k < sz && (w[k] -= log) <= 0) {
      if (tryFactorThis(k)) return true ;
    }
    return false ;
  }

  final static double sqrt2 = Math.sqrt(2.0) ;

  private void again() {
    for (int k = 0 ; k < sz;k++) {
      {
        BigInteger x = v.add(BigInteger.valueOf (k)) ;
        BigInteger y = poly(x) ;
        int [] t = factorThis(y) ;
        if (t !=null) {
          System.err.println("x=" + x +", y=" + y) ;
          if (t[t.length-1] > B.intValue()) {
            System.err.print("*") ;
          }
          System.err.print(x + "," + w[k] + ":") ;
          for (int i = 0 ; i < t.length ; i++)
            System.err.print(" " + t[i]) ;
          System.err.println() ;
        }
      }
    }
  }

  int fillW(int delta,
                    short w[], int i0, int sz, short m0, short m1) {
    if (m0 == m1 || sz <= 1) {
      short lim = (short)(m0+delta) ;
      for (int k = 0 ; k < sz ; k++)
        w[i0++] = lim ;
      return lim ;
    } else {
      int sz2 = sz/2 ;
      short m = (short)(poly(v.add(BigInteger.valueOf(i0+sz2))).bitLength()) ;
      return Math.min(fillW(delta, w, i0, sz2, m0, m),
                      fillW(delta, w, i0 + sz2, sz-sz2, m, m1)) ;
    }
  }

  boolean zyva(BigInteger v, int id) {
    int len = base.length ;
    BigInteger m = poly(v) ;
    short log1 = (short)m.bitLength() ;
    short log2 = (short)poly(v.add(BigInteger.valueOf(sz))).bitLength() ;
    int logM = Math.min(log1, log2) ;
    this.v = v ;

    // Initialize w[i]'s to their limit values
    short sieveLim =
      (short)fillW((short)-(logB), w, 0, sz, log1, log2) ;

    // Do not sieve small primes
    int log = 1 ;
    int lim = 3 ;
    int i ;

    for (i = 2 ; log < S ; i++) {
      int p = base[i] ;
      while (p >= lim) {
        log++ ;
        long tlog = log ;
        long k = (long)1 << tlog ;
        double x = k * sqrt2 ;
        lim = (int)x ;
      }
    }

    // Just sieve, do not check factors as long as log(p) < logLim
    // May miss a few equations, in practice very few
    if (verbose > 1) System.err.println("Start sieving: " + base[i]) ;
    if (len > 50) {
      int logLim = (8*logMaxPrime)/10 ;
    outer:
      for ( ;  ; i++) {
        int p = base[i] ;
      
        while (p >= lim) {
          log++ ;
          long tlog = log ;
          long k = (long)1 << tlog ;
          double x = k * sqrt2 ;
          lim = (int)x ;
          if (log >= logLim) break outer;
        }
        {
          int r = root[i] ;

          int vmod = v.remainder(bbase[i]).intValue() ;
          int xi=0 ,yi = 0;

          if (vmod < r) {
            xi = r-vmod ;
          } else if (vmod > r) {
            xi = (r+p) - vmod ;
          }
          r = p-r ;
          if (vmod < r) {
            yi = r-vmod ;
          } else if (vmod > r) {
            yi = (r+p) - vmod ;
          }
          if (xi > yi) { int t = xi ; xi=yi ; yi=t ; }
          int k = xi ;
          for (int kk=yi ; kk < sz ; k += p, kk += p) {
            w[k] -= log ;
            w[kk] -= log ;
          }
          if (k < sz)
            w[k] -= log ;
        }
      }
    }

    // Now sieve and check factors
    if (verbose > 1)
      System.err.println("Start sieving and factoring: " + base[i]) ;
    for ( ; i < len ; i++) {
      int p = base[i] ;
      while (p >= lim) {
        log++ ;
        long tlog = log ;
        long k = (long)1 << tlog ;
        double x = k * sqrt2 ;
        lim = (int)x ;
      }
      
      int r = root[i] ;
      int vmod = (int)v.remainder(bbase[i]).intValue() ;

      int xi = 0, yi=0 ;
      if (vmod < r) {
        xi = r-vmod ;
      } else if (vmod > r) {
        xi = (r+p) - vmod ;
      }
      r = p-r ;
      if (vmod < r) {
        yi = r-vmod ;
      } else if (vmod > r) {
        yi = (r+p) - vmod ;
      }

      if (xi > yi) { int t = xi ; xi=yi ; yi=t ; }

      int k = xi ;
      for (int kk = yi ; kk < sz ; k += p, kk += p) {
        if ((w[k] -= log) <= 0) {
          if (tryFactorThis(k)) return true ;
        }
        if ((w[kk] -= log) <= 0) {
          if (tryFactorThis(kk)) return true ;
        }
      }
      if (k < sz && (w[k] -= log) <= 0) {
        if (tryFactorThis(k)) return true ;
      }
    }
    //    again() ;
    if (verbose > 0) efficiency(id) ;
    return false ;
  }
}
